/* (C)opyright 2007 Cédric Krier <ced@ced.homedns.org>
 * inspired from macbook-backlight by Ryan Lortie <desrt@desrt.ca>
 *          and pommed by Julien Blache <jb@jblache.org>
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <pci/pci.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>

#define PCI_ID_VENDOR_INTEL 0x8086
#define PCI_ID_PRODUCT_GMA950 0x27A2
#define REGISTER_OFFSET 0x00061254
#define INREG(addr) readl(memory+addr)
#define OUTREG(addr,val) writel(val, memory+addr)

/* static */

static char *memory = NULL;
static long address = 0;
static long length = 0;
static unsigned int BACKLIGHT_MAX;
static unsigned int BACKLIGHT_MIN;

static void
usage(void) {
	fprintf(stderr, "usage: macbook-backlight [-p] [-s <value>] [-v]\n");
	exit(EXIT_FAILURE);
}

static inline unsigned int
readl(const volatile void *addr)
{
	return *(volatile unsigned int*) addr;
}

static inline void
writel(unsigned int b, volatile void *addr)
{
	*(volatile unsigned int*) addr = b;
}

static void
address_probe(void) {
	struct pci_access *pacc = NULL;
	struct pci_dev *dev = NULL;

	pacc = pci_alloc();
	if(pacc == NULL) {
		fprintf(stderr, "error, cannot allocate pci\n");
		exit(EXIT_FAILURE);
	}
	pci_init(pacc);
	pci_scan_bus(pacc);

	for(dev = pacc->devices; dev; dev = dev->next) {
		pci_fill_info(dev, PCI_FILL_IDENT | PCI_FILL_BASES);
		if((dev->vendor_id == PCI_ID_VENDOR_INTEL)
			&& (dev->device_id == PCI_ID_PRODUCT_GMA950)) {
			address = dev->base_addr[0];
			length = dev->size[0];
		}
	}
	pci_cleanup(pacc);
}

static void
memory_map(void) {
	int fd;

	if((address == 0) || (length == 0)) {
		fprintf(stderr, "error, cannot find video card\n");
		exit(EXIT_FAILURE);
	}

	fd = open("/dev/mem", O_RDWR);
	if(fd < 0) {
		fprintf(stderr, "error, cannot open /dev/mem\n");
		exit(EXIT_FAILURE);
	}

	memory = mmap(NULL, length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, address);
	close(fd);
	if(memory == MAP_FAILED) {
		fprintf(stderr, "error, mmap failed: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
}

static unsigned long
backlight_get(void) {
	return (INREG(REGISTER_OFFSET) >> 1) & 0x7fff;
}

static unsigned long
backlight_get_max(void) {
	return (INREG(REGISTER_OFFSET) >> 17);
}

static void
backlight_set(long value) {
	BACKLIGHT_MAX = backlight_get_max();
	BACKLIGHT_MIN = 0x1f;

	if(value > BACKLIGHT_MAX)
		value = BACKLIGHT_MAX;
	if(value < BACKLIGHT_MIN)
		value = 0;
	OUTREG(REGISTER_OFFSET, (BACKLIGHT_MAX << 17) | (value << 1));
}

static long
value_get(const char *str) {
	long value;
	char *end;

	value = strtol(str, &end, 10);
	if(*end) {
		fprintf(stderr, "error, invalid value\n");
		exit(EXIT_FAILURE);
	}
	return value;
}

int
main (int argc, char *argv[]) {
	int print_value = 0;
	int set_value = 0;
	int i;
	long value;

	for(i = 1; i < argc; i++)
		if(!strncmp(argv[i], "-p", 3)) {
			print_value = 1;
		}
		else if(!strncmp(argv[i], "-s", 3)) {
			if(++i < argc) value = value_get(argv[i]);
			set_value = 1;
		}
		else if(!strncmp(argv[i], "-v", 3)) {
			fprintf(stderr, "macbook-backlight-"VERSION", (C)opyright MMVII Cédric Krier\n");
			exit(EXIT_FAILURE);
		}
		else
			usage();
	address_probe();
	memory_map();

	if(set_value) {
		backlight_set(value);
	}

	if(print_value)
		fprintf(stdout, "%ld\n", backlight_get());
}
